//
//	RemoteJdbcJdbcDriver.java
//
//	version 1.0  98/12/12
//	author  Mitsumasa IWATSUKI
//

package org.iwatsuki.rjj;

import java.io.*;
import java.math.BigDecimal;
import java.net.*;
import java.rmi.*;
import java.rmi.server.UnicastRemoteObject;
import java.sql.*;
import java.util.*;

public class RemoteJdbcJdbcDriver implements Driver {
	private static Hashtable remoteDrivers = new Hashtable();
	private static final String COMMUNICATION_ERROR_STATE = "08001";
	private static final String INVALID_URL_STATE = "2E000";

	static {
		try {
			DriverManager.registerDriver(new RemoteJdbcJdbcDriver());
		} catch (SQLException e) {
			throw new ExceptionInInitializerError(e);
		}
	}

	public Connection connect(String url, java.util.Properties info) throws SQLException {
		try {
			RemoteDriver rd = getRemoteDriver(url);
			return (rd != null) ? new ConnectionProxy(rd.connect(getSubURL(url), info)) : null;
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean acceptsURL(String url) throws SQLException {
		try {
			RemoteDriver rd = getRemoteDriver(url);
			return (rd != null) ? rd.acceptsURL(getSubURL(url)) : false;
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public DriverPropertyInfo[] getPropertyInfo(String url, java.util.Properties info) throws SQLException {
		try {
			RemoteDriver rd = getRemoteDriver(url);
			return (rd != null) ? rd.getPropertyInfo(getSubURL(url), info) : null;
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMajorVersion() {
		return 1;
	}

	public int getMinorVersion() {
		return 0;
	}

	public boolean jdbcCompliant() {
		return false;
	}

	private synchronized RemoteDriver getRemoteDriver(String url) throws SQLException {
		try {
			String rmiURL = getRmiURL(url);
			if (rmiURL == null)
				throw new MalformedURLException("Invalid url: " + url);

			String name = "";
			if (rmiURL.startsWith("//")) {
				int index = rmiURL.indexOf('/', 2);
				if (index == -1)
					rmiURL += '/';
				else
					name = rmiURL.substring(++index).trim();
			} else if (rmiURL.startsWith("/")) {
				name = rmiURL.substring(1).trim();
			} else {
				name = rmiURL;
			}

			if (name.equals("")) {
				name = "RemoteDriver";
				rmiURL += name;
			}

			RemoteDriver rd = (RemoteDriver)remoteDrivers.get(rmiURL);
			if (rd == null) {
				rd = (RemoteDriver)Naming.lookup(rmiURL);
				remoteDrivers.put(rmiURL, rd);
			}
			return rd;
		} catch (MalformedURLException e) {
			throw new SQLException(e.getMessage(), INVALID_URL_STATE);
		} catch (NotBoundException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	private String getRmiURL(String url) {
		String host = null;

		String subName;
		if (url != null && (subName = getSubName(url)) != null) {
			int index = subName.indexOf(';');
			if (index != -1)
				host = subName.substring(0, index).trim();
		}
		return host;
	}

	private String getSubURL(String url) {
		String subURL = null;

		String subName;
		if (url != null && (subName = getSubName(url)) != null) {
			int index = subName.indexOf(';');
			if (index != -1)
				subURL = subName.substring(++index).trim();
		}
		return subURL;
	}

	private String getSubName(String url) {
		String subName = null;
		if (url != null) {
			int index = url.indexOf(':');
			if (index != -1 && url.substring(0, index).trim().equals("jdbc")) {
				url = url.substring(++index);
				index = url.indexOf(':');
				if (index != -1 && url.substring(0, index).trim().equals("rjj"))
					subName = url.substring(++index).trim();
			}
		}
		return subName;
	}
}

class ConnectionProxy implements Connection {
	private RemoteConnection con;
	private static final String COMMUNICATION_ERROR_STATE = "08001";

	public ConnectionProxy(RemoteConnection con) {
		this.con = con;
	}

	//----------------------------------------------------------------------

	public Statement createStatement() throws SQLException {
		try {
			return new StatementProxy(con.createStatement());
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public PreparedStatement prepareStatement(String sql) throws SQLException {
		try {
			return new PreparedStatementProxy(con.prepareStatement(sql));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public CallableStatement prepareCall(String sql) throws SQLException {
		try {
			return new CallableStatementProxy(con.prepareCall(sql));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String nativeSQL(String sql) throws SQLException {
		try {
			return con.nativeSQL(sql);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setAutoCommit(boolean autoCommit) throws SQLException {
		try {
			con.setAutoCommit(autoCommit);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean getAutoCommit() throws SQLException {
		try {
			return con.getAutoCommit();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void commit() throws SQLException {
		try {
			con.commit();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void rollback() throws SQLException {
		try {
			con.rollback();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void close() throws SQLException {
		try {
			con.close();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean isClosed() throws SQLException {
		try {
			return con.isClosed();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	//======================================================================
	// Advanced features:

	public DatabaseMetaData getMetaData() throws SQLException {
		try {
			return new DatabaseMetaDataProxy(con.getMetaData());
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setReadOnly(boolean readOnly) throws SQLException {
		try {
			con.setReadOnly(readOnly);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean isReadOnly() throws SQLException {
		try {
			return con.isReadOnly();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setCatalog(String catalog) throws SQLException {
		try {
			con.setCatalog(catalog);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getCatalog() throws SQLException {
		try {
			return con.getCatalog();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setTransactionIsolation(int level) throws SQLException {
		try {
			con.setTransactionIsolation(level);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getTransactionIsolation() throws SQLException {
		try {
			return con.getTransactionIsolation();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public SQLWarning getWarnings() throws SQLException {
		try {
			return con.getWarnings();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void clearWarnings() throws SQLException {
		try {
			con.clearWarnings();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}
}

class StatementProxy implements Statement {
	private RemoteStatement stmt;
	private static final String COMMUNICATION_ERROR_STATE = "08001";

	public StatementProxy(RemoteStatement stmt) {
		this.stmt = stmt;
	}

	//----------------------------------------------------------------------

	public ResultSet executeQuery(String sql) throws SQLException {
		try {
			return new ResultSetProxy(stmt.executeQuery(sql));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int executeUpdate(String sql) throws SQLException {
		try {
			return stmt.executeUpdate(sql);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void close() throws SQLException {
		try {
			stmt.close();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	//----------------------------------------------------------------------

	public int getMaxFieldSize() throws SQLException {
		try {
			return stmt.getMaxFieldSize();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setMaxFieldSize(int max) throws SQLException {
		try {
			stmt.setMaxFieldSize(max);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxRows() throws SQLException {
		try {
			return stmt.getMaxRows();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setMaxRows(int max) throws SQLException {
		try {
			stmt.setMaxRows(max);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setEscapeProcessing(boolean enable) throws SQLException {
		try {
			stmt.setEscapeProcessing(enable);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getQueryTimeout() throws SQLException {
		try {
			return stmt.getQueryTimeout();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setQueryTimeout(int seconds) throws SQLException {
		try {
			stmt.setQueryTimeout(seconds);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void cancel() throws SQLException {
		try {
			stmt.cancel();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public SQLWarning getWarnings() throws SQLException {
		try {
			return stmt.getWarnings();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void clearWarnings() throws SQLException {
		try {
			stmt.clearWarnings();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setCursorName(String name) throws SQLException {
		try {
			stmt.setCursorName(name);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	//----------------------- Multiple Results --------------------------

	public boolean execute(String sql) throws SQLException {
		try {
			return stmt.execute(sql);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getResultSet() throws SQLException {
		try {
			return new ResultSetProxy(stmt.getResultSet());
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getUpdateCount() throws SQLException {
		try {
			return stmt.getUpdateCount();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean getMoreResults() throws SQLException {
		try {
			return stmt.getMoreResults();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}
}

class PreparedStatementProxy extends StatementProxy implements PreparedStatement {
	private RemotePreparedStatement pstmt;
	private static final String COMMUNICATION_ERROR_STATE = "08001";
	private static final String UNEXPECTED_ERROR_STATE = "58033";

	public PreparedStatementProxy(RemotePreparedStatement pstmt) {
		super(pstmt);
		this.pstmt = pstmt;
	}

	private byte[] readByteArray(InputStream in, int length) throws IOException {
		byte[] buf = new byte[length];
		int count = in.read(buf);
		if (count < length) {
			if (count < 0)
				count = 0;
			byte[] newbuf = new byte[count];
			System.arraycopy(buf, 0, newbuf, 0, count);
			buf = newbuf;
		}
		return buf;
	}

	//----------------------------------------------------------------------

	public ResultSet executeQuery() throws SQLException {
		try {
			return new ResultSetProxy(pstmt.executeQuery());
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int executeUpdate() throws SQLException {
		try {
			return pstmt.executeUpdate();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setNull(int parameterIndex, int sqlType) throws SQLException {
		try {
			pstmt.setNull(parameterIndex, sqlType);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setBoolean(int parameterIndex, boolean x) throws SQLException {
		try {
			pstmt.setBoolean(parameterIndex, x);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setByte(int parameterIndex, byte x) throws SQLException {
		try {
			pstmt.setByte(parameterIndex, x);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setShort(int parameterIndex, short x) throws SQLException {
		try {
			pstmt.setShort(parameterIndex, x);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setInt(int parameterIndex, int x) throws SQLException {
		try {
			pstmt.setInt(parameterIndex, x);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setLong(int parameterIndex, long x) throws SQLException {
		try {
			pstmt.setLong(parameterIndex, x);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setFloat(int parameterIndex, float x) throws SQLException {
		try {
			pstmt.setFloat(parameterIndex, x);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setDouble(int parameterIndex, double x) throws SQLException {
		try {
			pstmt.setDouble(parameterIndex, x);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setBigDecimal(int parameterIndex, BigDecimal x) throws SQLException {
		try {
			pstmt.setBigDecimal(parameterIndex, x);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setString(int parameterIndex, String x) throws SQLException {
		try {
			pstmt.setString(parameterIndex, x);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setBytes(int parameterIndex, byte x[]) throws SQLException {
		try {
			pstmt.setBytes(parameterIndex, x);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setDate(int parameterIndex, java.sql.Date x) throws SQLException {
		try {
			pstmt.setDate(parameterIndex, x);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setTime(int parameterIndex, java.sql.Time x) throws SQLException {
		try {
			pstmt.setTime(parameterIndex, x);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setTimestamp(int parameterIndex, java.sql.Timestamp x) throws SQLException {
		try {
			pstmt.setTimestamp(parameterIndex, x);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setAsciiStream(int parameterIndex, java.io.InputStream x, int length) throws SQLException {
		try {
			pstmt.setAsciiStream(parameterIndex, readByteArray(x, length), length);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		} catch (IOException e) {
			throw new SQLException(e.getMessage(), UNEXPECTED_ERROR_STATE);
		}
	}

	public void setUnicodeStream(int parameterIndex, java.io.InputStream x, int length) throws SQLException {
		try {
			pstmt.setUnicodeStream(parameterIndex, readByteArray(x, length), length);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		} catch (IOException e) {
			throw new SQLException(e.getMessage(), UNEXPECTED_ERROR_STATE);
		}
	}

	public void setBinaryStream(int parameterIndex, java.io.InputStream x, int length) throws SQLException {
		try {
			pstmt.setBinaryStream(parameterIndex, readByteArray(x, length), length);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		} catch (IOException e) {
			throw new SQLException(e.getMessage(), UNEXPECTED_ERROR_STATE);
		}
	}

	public void clearParameters() throws SQLException {
		try {
			pstmt.clearParameters();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	//----------------------------------------------------------------------
	// Advanced features:

	public void setObject(int parameterIndex, Object x, int targetSqlType, int scale) throws SQLException {
		try {
			pstmt.setObject(parameterIndex, x, targetSqlType, scale);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setObject(int parameterIndex, Object x, int targetSqlType) throws SQLException {
		try {
			pstmt.setObject(parameterIndex, x, targetSqlType);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void setObject(int parameterIndex, Object x) throws SQLException {
		try {
			pstmt.setObject(parameterIndex, x);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean execute() throws SQLException {
		try {
			return pstmt.execute();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}
}

class CallableStatementProxy extends PreparedStatementProxy implements CallableStatement {
	private RemoteCallableStatement cstmt;
	private static final String COMMUNICATION_ERROR_STATE = "08001";

	public CallableStatementProxy(RemoteCallableStatement cstmt) {
		super(cstmt);
		this.cstmt = cstmt;
	}

	//----------------------------------------------------------------------

	public void registerOutParameter(int parameterIndex, int sqlType) throws SQLException {
		try {
			cstmt.registerOutParameter(parameterIndex, sqlType);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void registerOutParameter(int parameterIndex, int sqlType, int scale) throws SQLException {
		try {
			cstmt.registerOutParameter(parameterIndex, sqlType, scale);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean wasNull() throws SQLException {
		try {
			return cstmt.wasNull();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getString(int parameterIndex) throws SQLException {
		try {
			return cstmt.getString(parameterIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean getBoolean(int parameterIndex) throws SQLException {
		try {
			return cstmt.getBoolean(parameterIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public byte getByte(int parameterIndex) throws SQLException {
		try {
			return cstmt.getByte(parameterIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public short getShort(int parameterIndex) throws SQLException {
		try {
			return cstmt.getShort(parameterIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getInt(int parameterIndex) throws SQLException {
		try {
			return cstmt.getInt(parameterIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public long getLong(int parameterIndex) throws SQLException {
		try {
			return cstmt.getLong(parameterIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public float getFloat(int parameterIndex) throws SQLException {
		try {
			return cstmt.getFloat(parameterIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public double getDouble(int parameterIndex) throws SQLException {
		try {
			return cstmt.getDouble(parameterIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public BigDecimal getBigDecimal(int parameterIndex, int scale) throws SQLException {
		try {
			return cstmt.getBigDecimal(parameterIndex, scale);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public byte[] getBytes(int parameterIndex) throws SQLException {
		try {
			return cstmt.getBytes(parameterIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.sql.Date getDate(int parameterIndex) throws SQLException {
		try {
			return cstmt.getDate(parameterIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.sql.Time getTime(int parameterIndex) throws SQLException {
		try {
			return cstmt.getTime(parameterIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.sql.Timestamp getTimestamp(int parameterIndex) throws SQLException {
		try {
			return cstmt.getTimestamp(parameterIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	//----------------------------------------------------------------------
	// Advanced features:

	public Object getObject(int parameterIndex) throws SQLException {
		try {
			return cstmt.getObject(parameterIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}
}

class ResultSetProxy implements ResultSet {
	private RemoteResultSet rs;
	private static final String COMMUNICATION_ERROR_STATE = "08001";

	public ResultSetProxy(RemoteResultSet rs) {
		this.rs = rs;
	}

	//----------------------------------------------------------------------

	public boolean next() throws SQLException {
		try {
			return rs.next();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void close() throws SQLException {
		try {
			rs.close();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean wasNull() throws SQLException {
		try {
			return rs.wasNull();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	//======================================================================
	// Methods for accessing results by column index
	//======================================================================

	public String getString(int columnIndex) throws SQLException {
		try {
			return rs.getString(columnIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean getBoolean(int columnIndex) throws SQLException {
		try {
			return rs.getBoolean(columnIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public byte getByte(int columnIndex) throws SQLException {
		try {
			return rs.getByte(columnIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public short getShort(int columnIndex) throws SQLException {
		try {
			return rs.getShort(columnIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getInt(int columnIndex) throws SQLException {
		try {
			return rs.getInt(columnIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public long getLong(int columnIndex) throws SQLException {
		try {
			return rs.getLong(columnIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public float getFloat(int columnIndex) throws SQLException {
		try {
			return rs.getFloat(columnIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public double getDouble(int columnIndex) throws SQLException {
		try {
			return rs.getDouble(columnIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public BigDecimal getBigDecimal(int columnIndex, int scale) throws SQLException {
		try {
			return rs.getBigDecimal(columnIndex, scale);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public byte[] getBytes(int columnIndex) throws SQLException {
		try {
			return rs.getBytes(columnIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.sql.Date getDate(int columnIndex) throws SQLException {
		try {
			return rs.getDate(columnIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.sql.Time getTime(int columnIndex) throws SQLException {
		try {
			return rs.getTime(columnIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.sql.Timestamp getTimestamp(int columnIndex) throws SQLException {
		try {
			return rs.getTimestamp(columnIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.io.InputStream getAsciiStream(int columnIndex) throws SQLException {
		try {
			return new InputStreamProxy(rs.getAsciiStream(columnIndex));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.io.InputStream getUnicodeStream(int columnIndex) throws SQLException {
		try {
			return new InputStreamProxy(rs.getUnicodeStream(columnIndex));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.io.InputStream getBinaryStream(int columnIndex) throws SQLException {
		try {
			return new InputStreamProxy(rs.getBinaryStream(columnIndex));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	//======================================================================
	// Methods for accessing results by column name
	//======================================================================

	public String getString(String columnName) throws SQLException {
		try {
			return rs.getString(columnName);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean getBoolean(String columnName) throws SQLException {
		try {
			return rs.getBoolean(columnName);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public byte getByte(String columnName) throws SQLException {
		try {
			return rs.getByte(columnName);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public short getShort(String columnName) throws SQLException {
		try {
			return rs.getShort(columnName);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getInt(String columnName) throws SQLException {
		try {
			return rs.getInt(columnName);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public long getLong(String columnName) throws SQLException {
		try {
			return rs.getLong(columnName);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public float getFloat(String columnName) throws SQLException {
		try {
			return rs.getFloat(columnName);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public double getDouble(String columnName) throws SQLException {
		try {
			return rs.getDouble(columnName);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public BigDecimal getBigDecimal(String columnName, int scale) throws SQLException {
		try {
			return rs.getBigDecimal(columnName, scale);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public byte[] getBytes(String columnName) throws SQLException {
		try {
			return rs.getBytes(columnName);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.sql.Date getDate(String columnName) throws SQLException {
		try {
			return rs.getDate(columnName);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.sql.Time getTime(String columnName) throws SQLException {
		try {
			return rs.getTime(columnName);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.sql.Timestamp getTimestamp(String columnName) throws SQLException {
		try {
			return rs.getTimestamp(columnName);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.io.InputStream getAsciiStream(String columnName) throws SQLException {
		try {
			return new InputStreamProxy(rs.getAsciiStream(columnName));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.io.InputStream getUnicodeStream(String columnName) throws SQLException {
		try {
			return new InputStreamProxy(rs.getUnicodeStream(columnName));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public java.io.InputStream getBinaryStream(String columnName) throws SQLException {
		try {
			return new InputStreamProxy(rs.getBinaryStream(columnName));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	//=====================================================================
	// Advanced features:
	//=====================================================================

	public SQLWarning getWarnings() throws SQLException {
		try {
			return rs.getWarnings();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public void clearWarnings() throws SQLException {
		try {
			rs.clearWarnings();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getCursorName() throws SQLException {
		try {
			return rs.getCursorName();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSetMetaData getMetaData() throws SQLException {
		try {
			return new ResultSetMetaDataProxy(rs.getMetaData());
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public Object getObject(int columnIndex) throws SQLException {
		try {
			return rs.getObject(columnIndex);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public Object getObject(String columnName) throws SQLException {
		try {
			return rs.getObject(columnName);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	//----------------------------------------------------------------

	public int findColumn(String columnName) throws SQLException {
		try {
			return rs.findColumn(columnName);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}
}

class DatabaseMetaDataProxy implements DatabaseMetaData {
	private RemoteDatabaseMetaData dm;
	private static final String COMMUNICATION_ERROR_STATE = "08001";

	public DatabaseMetaDataProxy(RemoteDatabaseMetaData dm) {
		this.dm = dm;
	}

	//----------------------------------------------------------------------
	// First, a variety of minor information about the target database.

	public boolean allProceduresAreCallable() throws SQLException {
		try {
			return dm.allProceduresAreCallable();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean allTablesAreSelectable() throws SQLException {
		try {
			return dm.allTablesAreSelectable();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getURL() throws SQLException {
		try {
			return dm.getURL();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getUserName() throws SQLException {
		try {
			return dm.getUserName();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean isReadOnly() throws SQLException {
		try {
			return dm.isReadOnly();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean nullsAreSortedHigh() throws SQLException {
		try {
			return dm.nullsAreSortedHigh();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean nullsAreSortedLow() throws SQLException {
		try {
			return dm.nullsAreSortedLow();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean nullsAreSortedAtStart() throws SQLException {
		try {
			return dm.nullsAreSortedAtStart();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean nullsAreSortedAtEnd() throws SQLException {
		try {
			return dm.nullsAreSortedAtEnd();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getDatabaseProductName() throws SQLException {
		try {
			return dm.getDatabaseProductName();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getDatabaseProductVersion() throws SQLException {
		try {
			return dm.getDatabaseProductVersion();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getDriverName() throws SQLException {
		try {
			return dm.getDriverName();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getDriverVersion() throws SQLException {
		try {
			return dm.getDriverVersion();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getDriverMajorVersion() {
		try {
			return dm.getDriverMajorVersion();
		} catch (RemoteException e) {
			return 0;
		}
	}

	public int getDriverMinorVersion() {
		try {
			return dm.getDriverMinorVersion();
		} catch (RemoteException e) {
			return 0;
		}
	}

	public boolean usesLocalFiles() throws SQLException {
		try {
			return dm.usesLocalFiles();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean usesLocalFilePerTable() throws SQLException {
		try {
			return dm.usesLocalFilePerTable();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsMixedCaseIdentifiers() throws SQLException {
		try {
			return dm.supportsMixedCaseIdentifiers();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean storesUpperCaseIdentifiers() throws SQLException {
		try {
			return dm.storesUpperCaseIdentifiers();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean storesLowerCaseIdentifiers() throws SQLException {
		try {
			return dm.storesLowerCaseIdentifiers();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean storesMixedCaseIdentifiers() throws SQLException {
		try {
			return dm.storesMixedCaseIdentifiers();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsMixedCaseQuotedIdentifiers() throws SQLException {
		try {
			return dm.supportsMixedCaseQuotedIdentifiers();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean storesUpperCaseQuotedIdentifiers() throws SQLException {
		try {
			return dm.storesUpperCaseQuotedIdentifiers();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean storesLowerCaseQuotedIdentifiers() throws SQLException {
		try {
			return dm.storesLowerCaseQuotedIdentifiers();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean storesMixedCaseQuotedIdentifiers() throws SQLException {
		try {
			return dm.storesMixedCaseQuotedIdentifiers();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getIdentifierQuoteString() throws SQLException {
		try {
			return dm.getIdentifierQuoteString();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getSQLKeywords() throws SQLException {
		try {
			return dm.getSQLKeywords();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getNumericFunctions() throws SQLException {
		try {
			return dm.getNumericFunctions();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getStringFunctions() throws SQLException {
		try {
			return dm.getStringFunctions();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getSystemFunctions() throws SQLException {
		try {
			return dm.getSystemFunctions();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getTimeDateFunctions() throws SQLException {
		try {
			return dm.getTimeDateFunctions();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getSearchStringEscape() throws SQLException {
		try {
			return dm.getSearchStringEscape();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getExtraNameCharacters() throws SQLException {
		try {
			return dm.getExtraNameCharacters();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	//--------------------------------------------------------------------
	// Functions describing which features are supported.

	public boolean supportsAlterTableWithAddColumn() throws SQLException {
		try {
			return dm.supportsAlterTableWithAddColumn();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsAlterTableWithDropColumn() throws SQLException {
		try {
			return dm.supportsAlterTableWithDropColumn();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsColumnAliasing() throws SQLException {
		try {
			return dm.supportsColumnAliasing();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean nullPlusNonNullIsNull() throws SQLException {
		try {
			return dm.nullPlusNonNullIsNull();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsConvert() throws SQLException {
		try {
			return dm.supportsConvert();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsConvert(int fromType, int toType) throws SQLException {
		try {
			return dm.supportsConvert(fromType, toType);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsTableCorrelationNames() throws SQLException {
		try {
			return dm.supportsTableCorrelationNames();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsDifferentTableCorrelationNames() throws SQLException {
		try {
			return dm.supportsDifferentTableCorrelationNames();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsExpressionsInOrderBy() throws SQLException {
		try {
			return dm.supportsExpressionsInOrderBy();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsOrderByUnrelated() throws SQLException {
		try {
			return dm.supportsOrderByUnrelated();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsGroupBy() throws SQLException {
		try {
			return dm.supportsGroupBy();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsGroupByUnrelated() throws SQLException {
		try {
			return dm.supportsGroupByUnrelated();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsGroupByBeyondSelect() throws SQLException {
		try {
			return dm.supportsGroupByBeyondSelect();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsLikeEscapeClause() throws SQLException {
		try {
			return dm.supportsLikeEscapeClause();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsMultipleResultSets() throws SQLException {
		try {
			return dm.supportsMultipleResultSets();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsMultipleTransactions() throws SQLException {
		try {
			return dm.supportsMultipleTransactions();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsNonNullableColumns() throws SQLException {
		try {
			return dm.supportsNonNullableColumns();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsMinimumSQLGrammar() throws SQLException {
		try {
			return dm.supportsMinimumSQLGrammar();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsCoreSQLGrammar() throws SQLException {
		try {
			return dm.supportsCoreSQLGrammar();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsExtendedSQLGrammar() throws SQLException {
		try {
			return dm.supportsExtendedSQLGrammar();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsANSI92EntryLevelSQL() throws SQLException {
		try {
			return dm.supportsANSI92EntryLevelSQL();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsANSI92IntermediateSQL() throws SQLException {
		try {
			return dm.supportsANSI92IntermediateSQL();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsANSI92FullSQL() throws SQLException {
		try {
			return dm.supportsANSI92FullSQL();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsIntegrityEnhancementFacility() throws SQLException {
		try {
			return dm.supportsIntegrityEnhancementFacility();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsOuterJoins() throws SQLException {
		try {
			return dm.supportsOuterJoins();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsFullOuterJoins() throws SQLException {
		try {
			return dm.supportsFullOuterJoins();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsLimitedOuterJoins() throws SQLException {
		try {
			return dm.supportsLimitedOuterJoins();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getSchemaTerm() throws SQLException {
		try {
			return dm.getSchemaTerm();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getProcedureTerm() throws SQLException {
		try {
			return dm.getProcedureTerm();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getCatalogTerm() throws SQLException {
		try {
			return dm.getCatalogTerm();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean isCatalogAtStart() throws SQLException {
		try {
			return dm.isCatalogAtStart();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getCatalogSeparator() throws SQLException {
		try {
			return dm.getCatalogSeparator();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsSchemasInDataManipulation() throws SQLException {
		try {
			return dm.supportsSchemasInDataManipulation();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsSchemasInProcedureCalls() throws SQLException {
		try {
			return dm.supportsSchemasInProcedureCalls();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsSchemasInTableDefinitions() throws SQLException {
		try {
			return dm.supportsSchemasInTableDefinitions();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsSchemasInIndexDefinitions() throws SQLException {
		try {
			return dm.supportsSchemasInIndexDefinitions();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsSchemasInPrivilegeDefinitions() throws SQLException {
		try {
			return dm.supportsSchemasInPrivilegeDefinitions();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsCatalogsInDataManipulation() throws SQLException {
		try {
			return dm.supportsCatalogsInDataManipulation();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsCatalogsInProcedureCalls() throws SQLException {
		try {
			return dm.supportsCatalogsInProcedureCalls();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsCatalogsInTableDefinitions() throws SQLException {
		try {
			return dm.supportsCatalogsInTableDefinitions();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsCatalogsInIndexDefinitions() throws SQLException {
		try {
			return dm.supportsCatalogsInIndexDefinitions();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsCatalogsInPrivilegeDefinitions() throws SQLException {
		try {
			return dm.supportsCatalogsInPrivilegeDefinitions();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsPositionedDelete() throws SQLException {
		try {
			return dm.supportsPositionedDelete();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsPositionedUpdate() throws SQLException {
		try {
			return dm.supportsPositionedUpdate();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsSelectForUpdate() throws SQLException {
		try {
			return dm.supportsSelectForUpdate();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsStoredProcedures() throws SQLException {
		try {
			return dm.supportsStoredProcedures();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsSubqueriesInComparisons() throws SQLException {
		try {
			return dm.supportsSubqueriesInComparisons();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsSubqueriesInExists() throws SQLException {
		try {
			return dm.supportsSubqueriesInExists();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsSubqueriesInIns() throws SQLException {
		try {
			return dm.supportsSubqueriesInIns();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsSubqueriesInQuantifieds() throws SQLException {
		try {
			return dm.supportsSubqueriesInQuantifieds();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsCorrelatedSubqueries() throws SQLException {
		try {
			return dm.supportsCorrelatedSubqueries();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsUnion() throws SQLException {
		try {
			return dm.supportsUnion();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsUnionAll() throws SQLException {
		try {
			return dm.supportsUnionAll();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsOpenCursorsAcrossCommit() throws SQLException {
		try {
			return dm.supportsOpenCursorsAcrossCommit();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsOpenCursorsAcrossRollback() throws SQLException {
		try {
			return dm.supportsOpenCursorsAcrossRollback();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsOpenStatementsAcrossCommit() throws SQLException {
		try {
			return dm.supportsOpenStatementsAcrossCommit();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsOpenStatementsAcrossRollback() throws SQLException {
		try {
			return dm.supportsOpenStatementsAcrossRollback();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	//----------------------------------------------------------------------
	// The following group of methods exposes various limitations
	// based on the target database with the current driver.
	// Unless otherwise specified, a result of zero means there is no
	// limit, or the limit is not known.

	public int getMaxBinaryLiteralLength() throws SQLException {
		try {
			return dm.getMaxBinaryLiteralLength();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxCharLiteralLength() throws SQLException {
		try {
			return dm.getMaxCharLiteralLength();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxColumnNameLength() throws SQLException {
		try {
			return dm.getMaxColumnNameLength();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxColumnsInGroupBy() throws SQLException {
		try {
			return dm.getMaxColumnsInGroupBy();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxColumnsInIndex() throws SQLException {
		try {
			return dm.getMaxColumnsInIndex();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxColumnsInOrderBy() throws SQLException {
		try {
			return dm.getMaxColumnsInOrderBy();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxColumnsInSelect() throws SQLException {
		try {
			return dm.getMaxColumnsInSelect();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxColumnsInTable() throws SQLException {
		try {
			return dm.getMaxColumnsInTable();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxConnections() throws SQLException {
		try {
			return dm.getMaxConnections();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxCursorNameLength() throws SQLException {
		try {
			return dm.getMaxCursorNameLength();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxIndexLength() throws SQLException {
		try {
			return dm.getMaxIndexLength();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxSchemaNameLength() throws SQLException {
		try {
			return dm.getMaxSchemaNameLength();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxProcedureNameLength() throws SQLException {
		try {
			return dm.getMaxProcedureNameLength();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxCatalogNameLength() throws SQLException {
		try {
			return dm.getMaxCatalogNameLength();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxRowSize() throws SQLException {
		try {
			return dm.getMaxRowSize();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean doesMaxRowSizeIncludeBlobs() throws SQLException {
		try {
			return dm.doesMaxRowSizeIncludeBlobs();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxStatementLength() throws SQLException {
		try {
			return dm.getMaxStatementLength();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxStatements() throws SQLException {
		try {
			return dm.getMaxStatements();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxTableNameLength() throws SQLException {
		try {
			return dm.getMaxTableNameLength();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxTablesInSelect() throws SQLException {
		try {
			return dm.getMaxTablesInSelect();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getMaxUserNameLength() throws SQLException {
		try {
			return dm.getMaxUserNameLength();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	//----------------------------------------------------------------------

	public int getDefaultTransactionIsolation() throws SQLException {
		try {
			return dm.getDefaultTransactionIsolation();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsTransactions() throws SQLException {
		try {
			return dm.supportsTransactions();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsTransactionIsolationLevel(int level) throws SQLException {
		try {
			return dm.supportsTransactionIsolationLevel(level);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsDataDefinitionAndDataManipulationTransactions() throws SQLException {
		try {
			return dm.supportsDataDefinitionAndDataManipulationTransactions();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean supportsDataManipulationTransactionsOnly() throws SQLException {
		try {
			return dm.supportsDataManipulationTransactionsOnly();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean dataDefinitionCausesTransactionCommit() throws SQLException {
		try {
			return dm.dataDefinitionCausesTransactionCommit();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean dataDefinitionIgnoredInTransactions() throws SQLException {
		try {
			return dm.dataDefinitionIgnoredInTransactions();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getProcedures(String catalog, String schemaPattern, String procedureNamePattern) throws SQLException {
		try {
			return new ResultSetProxy(dm.getProcedures(catalog, schemaPattern, procedureNamePattern));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getProcedureColumns(String catalog, String schemaPattern, String procedureNamePattern, String columnNamePattern) throws SQLException {
		try {
			return new ResultSetProxy(dm.getProcedureColumns(catalog, schemaPattern, procedureNamePattern, columnNamePattern));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getTables(String catalog, String schemaPattern, String tableNamePattern, String types[]) throws SQLException {
		try {
			return new ResultSetProxy(dm.getTables(catalog, schemaPattern, tableNamePattern, types));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getSchemas() throws SQLException {
		try {
			return new ResultSetProxy(dm.getSchemas());
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getCatalogs() throws SQLException {
		try {
			return new ResultSetProxy(dm.getCatalogs());
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getTableTypes() throws SQLException {
		try {
			return new ResultSetProxy(dm.getTableTypes());
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getColumns(String catalog, String schemaPattern, String tableNamePattern, String columnNamePattern) throws SQLException {
		try {
			return new ResultSetProxy(dm.getColumns(catalog, schemaPattern, tableNamePattern, columnNamePattern));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getColumnPrivileges(String catalog, String schema, String table, String columnNamePattern) throws SQLException {
		try {
			return new ResultSetProxy(dm.getColumnPrivileges(catalog, schema, table, columnNamePattern));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getTablePrivileges(String catalog, String schemaPattern, String tableNamePattern) throws SQLException {
		try {
			return new ResultSetProxy(dm.getTablePrivileges(catalog, schemaPattern, tableNamePattern));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getBestRowIdentifier(String catalog, String schema, String table, int scope, boolean nullable) throws SQLException {
		try {
			return new ResultSetProxy(dm.getBestRowIdentifier(catalog, schema, table, scope, nullable));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getVersionColumns(String catalog, String schema, String table) throws SQLException {
		try {
			return new ResultSetProxy(dm.getVersionColumns(catalog, schema, table));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getPrimaryKeys(String catalog, String schema, String table) throws SQLException {
		try {
			return new ResultSetProxy(dm.getPrimaryKeys(catalog, schema, table));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getImportedKeys(String catalog, String schema, String table) throws SQLException {
		try {
			return new ResultSetProxy(dm.getImportedKeys(catalog, schema, table));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getExportedKeys(String catalog, String schema, String table) throws SQLException {
		try {
			return new ResultSetProxy(dm.getExportedKeys(catalog, schema, table));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getCrossReference(String primaryCatalog, String primarySchema, String primaryTable, String foreignCatalog, String foreignSchema, String foreignTable) throws SQLException {
		try {
			return new ResultSetProxy(dm.getCrossReference(primaryCatalog, primarySchema, primaryTable, foreignCatalog, foreignSchema, foreignTable));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getTypeInfo() throws SQLException {
		try {
			return new ResultSetProxy(dm.getTypeInfo());
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public ResultSet getIndexInfo(String catalog, String schema, String table, boolean unique, boolean approximate) throws SQLException {
		try {
			return new ResultSetProxy(dm.getIndexInfo(catalog, schema, table, unique, approximate));
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}
}

class ResultSetMetaDataProxy implements ResultSetMetaData {
	private RemoteResultSetMetaData rsmd;
	private static final String COMMUNICATION_ERROR_STATE = "08001";

	public ResultSetMetaDataProxy(RemoteResultSetMetaData rsmd) {
		this.rsmd = rsmd;
	}

	//----------------------------------------------------------------------

	public int getColumnCount() throws SQLException {
		try {
			return rsmd.getColumnCount();
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean isAutoIncrement(int column) throws SQLException {
		try {
			return rsmd.isAutoIncrement(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean isCaseSensitive(int column) throws SQLException {
		try {
			return rsmd.isCaseSensitive(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean isSearchable(int column) throws SQLException {
		try {
			return rsmd.isSearchable(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean isCurrency(int column) throws SQLException {
		try {
			return rsmd.isCurrency(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int isNullable(int column) throws SQLException {
		try {
			return rsmd.isNullable(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean isSigned(int column) throws SQLException {
		try {
			return rsmd.isSigned(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getColumnDisplaySize(int column) throws SQLException {
		try {
			return rsmd.getColumnDisplaySize(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getColumnLabel(int column) throws SQLException {
		try {
			return rsmd.getColumnLabel(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getColumnName(int column) throws SQLException {
		try {
			return rsmd.getColumnName(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getSchemaName(int column) throws SQLException {
		try {
			return rsmd.getSchemaName(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getPrecision(int column) throws SQLException {
		try {
			return rsmd.getPrecision(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getScale(int column) throws SQLException {
		try {
			return rsmd.getScale(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getTableName(int column) throws SQLException {
		try {
			return rsmd.getTableName(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getCatalogName(int column) throws SQLException {
		try {
			return rsmd.getCatalogName(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public int getColumnType(int column) throws SQLException {
		try {
			return rsmd.getColumnType(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public String getColumnTypeName(int column) throws SQLException {
		try {
			return rsmd.getColumnTypeName(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean isReadOnly(int column) throws SQLException {
		try {
			return rsmd.isReadOnly(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean isWritable(int column) throws SQLException {
		try {
			return rsmd.isWritable(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}

	public boolean isDefinitelyWritable(int column) throws SQLException {
		try {
			return rsmd.isDefinitelyWritable(column);
		} catch (RemoteException e) {
			throw new SQLException(e.getMessage(), COMMUNICATION_ERROR_STATE);
		}
	}
}

class InputStreamProxy extends InputStream {
	private RemoteInputStream in;

	public InputStreamProxy(RemoteInputStream in) {
		this.in = in;
	}

	//----------------------------------------------------------------------

	public int read() throws IOException {
		try {
			return in.read();
		} catch (RemoteException e) {
			throw new IOException(e.getMessage());
		}
	}

	public int read(byte b[]) throws IOException {
		try {
			return in.read(b);
		} catch (RemoteException e) {
			throw new IOException(e.getMessage());
		}
	}

	public int read(byte b[], int off, int len) throws IOException {
		try {
			return in.read(b, off, len);
		} catch (RemoteException e) {
			throw new IOException(e.getMessage());
		}
	}

	public long skip(long n) throws IOException {
		try {
			return in.skip(n);
		} catch (RemoteException e) {
			throw new IOException(e.getMessage());
		}
	}

	public int available() throws IOException {
		try {
			return in.available();
		} catch (RemoteException e) {
			throw new IOException(e.getMessage());
		}
	}

	public void close() throws IOException {
		try {
			in.close();
		} catch (RemoteException e) {
			throw new IOException(e.getMessage());
		}
	}
}
